package net.knarcraft.placeholdersigns.handler;

import net.knarcraft.placeholdersigns.container.PlaceholderSign;
import net.knarcraft.placeholdersigns.container.QueuedPlaceholderSign;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.PriorityQueue;

/**
 * A handler for dealing with the placeholder sign update queue
 */
public class PlaceholderSignUpdateQueueHandler {

    private final PriorityQueue<QueuedPlaceholderSign> signUpdateQueue;

    /**
     * Instantiates a new placeholder sign update queue handler
     */
    public PlaceholderSignUpdateQueueHandler() {
        this.signUpdateQueue = new PriorityQueue<>();
    }

    /**
     * Polls the queue for any placeholder signs due for update
     *
     * <p>No placeholder sign will be returned unless they are due to be updated.</p>
     *
     * @return <p>The next queued placeholder sign, or null if no such placeholder sign exists</p>
     */
    @Nullable
    public PlaceholderSign pollQueue() {
        QueuedPlaceholderSign nextSign = signUpdateQueue.peek();
        if (nextSign == null) {
            return null;
        }

        long currentTime = System.currentTimeMillis();
        if (nextSign.updateTimestamp() < currentTime) {
            nextSign = signUpdateQueue.poll();
            if (nextSign != null) {
                return nextSign.placeholderSign();
            }
        }
        return null;
    }

    /**
     * Peeks at the next item in the queue
     *
     * @return <p>The next item in the queue, or null if empty</p>
     */
    @Nullable
    public QueuedPlaceholderSign peekQueue() {
        return signUpdateQueue.peek();
    }

    /**
     * Adds the specified queued sign to the queue
     *
     * @param queuedSign <p>The queued sign to queue</p>
     */
    public void queueSign(@NotNull QueuedPlaceholderSign queuedSign) {
        this.signUpdateQueue.add(queuedSign);
    }

    /**
     * '
     * Removes the specified placeholder sign from the queue
     *
     * @param placeholderSign <p>The placeholder sign to remove</p>
     */
    public void unQueueSign(@NotNull PlaceholderSign placeholderSign) {
        this.signUpdateQueue.removeIf((item) -> item.placeholderSign().equals(placeholderSign));
    }

}
