package net.knarcraft.placeholdersigns.listener;

import net.knarcraft.knarlib.formatting.StringFormatter;
import net.knarcraft.knarlib.property.ColorConversion;
import net.knarcraft.knarlib.util.ColorHelper;
import net.knarcraft.placeholdersigns.PlaceholderSigns;
import net.knarcraft.placeholdersigns.config.PlaceholderSignMessage;
import net.knarcraft.placeholdersigns.config.PlaceholderSignsPermission;
import net.knarcraft.placeholdersigns.container.PlaceholderSign;
import net.knarcraft.placeholdersigns.container.SignCopyRequest;
import net.knarcraft.placeholdersigns.container.SignLineChangeRequest;
import net.knarcraft.placeholdersigns.container.SignTextCopyRequest;
import net.knarcraft.placeholdersigns.handler.PlaceholderSignHandler;
import net.knarcraft.placeholdersigns.handler.PlaceholderSignRequestHandler;
import org.bukkit.Bukkit;
import org.bukkit.block.Sign;
import org.bukkit.block.sign.Side;
import org.bukkit.block.sign.SignSide;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Map;
import java.util.Objects;

import static net.knarcraft.placeholdersigns.util.SignSideHelper.getOpposite;
import static net.knarcraft.placeholdersigns.util.SignSideHelper.getSide;

/**
 * A listener for placeholder signs being clicked
 */
public class SignClickListener implements Listener {

    @EventHandler(priority = EventPriority.MONITOR)
    public void onSignClick(@NotNull PlayerInteractEvent event) {
        // Ignore if not a clicked sign
        if (!event.hasBlock() || event.getClickedBlock() == null ||
                !(event.getClickedBlock().getState() instanceof Sign sign)) {
            return;
        }

        Player player = event.getPlayer();
        PlaceholderSignRequestHandler requestHandler = PlaceholderSigns.getInstance().getRequestHandler();

        SignCopyRequest signCopyRequest = requestHandler.getSignCopyRequest(player);
        if (signCopyRequest != null) {
            if (checkWaxEdit(sign, player, PlaceholderSignsPermission.BYPASS_WAXED_COPY.getPermissionNode())) {
                return;
            }

            pasteSign(signCopyRequest.sign(), sign, player);
            event.setCancelled(true);
            return;
        }

        // Check if the player has run the /setSignLine command
        SignLineChangeRequest signChangeRequest = requestHandler.getSignChangeRequest(player);
        if (signChangeRequest != null) {
            if (checkWaxEdit(sign, player, PlaceholderSignsPermission.BYPASS_WAXED_EDIT.getPermissionNode())) {
                return;
            }

            doSignChange(sign, signChangeRequest.line(), player, getSide(sign, player), signChangeRequest.text());
            event.setCancelled(true);
            return;
        }

        SignTextCopyRequest signTextCopyRequest = requestHandler.getSignTextCopyRequest(player);
        if (signTextCopyRequest != null) {
            if (checkWaxEdit(sign, player, PlaceholderSignsPermission.BYPASS_WAXED_COPY.getPermissionNode())) {
                return;
            }

            pasteSignText(sign, signTextCopyRequest, player, getSide(sign, player));
            event.setCancelled(true);
        }
    }

    /**
     * Pastes sign text according to the given signTextCopyRequest
     *
     * @param targetSign          <p>The sign to paste onto</p>
     * @param signTextCopyRequest <p>The requested sign text copy request to fulfill</p>
     * @param player              <p>The player wanting to paste</p>
     * @param clickedSide         <p>The sign side the player wants to paste onto</p>
     */
    private void pasteSignText(@NotNull Sign targetSign, @NotNull SignTextCopyRequest signTextCopyRequest,
                               @NotNull Player player, @NotNull Side clickedSide) {
        Side sourceSide = signTextCopyRequest.side();
        Sign sourceSign = signTextCopyRequest.sign();
        PlaceholderSignHandler signHandler = PlaceholderSigns.getInstance().getSignHandler();
        PlaceholderSign sourcePlaceholderSign = signHandler.getFromLocation(sourceSign.getLocation());
        PlaceholderSign targetPlaceholderSign = signHandler.getFromLocation(targetSign.getLocation());

        if (sourceSide == null) {
            // Copy both sides
            copySignText(signHandler, sourcePlaceholderSign, targetPlaceholderSign, clickedSide, sourceSign, targetSign,
                    player);
        } else if (signTextCopyRequest.sourceLine() == null) {
            // Copy the side the player looked at while executing the command
            copySignSideText(signHandler, sourcePlaceholderSign, targetPlaceholderSign, sourceSide, clickedSide,
                    sourceSign, targetSign, player);
        } else {
            copySignLine(signTextCopyRequest, sourcePlaceholderSign, sourceSide, clickedSide, targetSign, player);
        }
    }

    /**
     * Copies sign line from one sign side to another
     *
     * @param signTextCopyRequest   <p>The sign text copy request to handle</p>
     * @param sourcePlaceholderSign <p>The source sign's placeholder sign, or null</p>
     * @param sourceSide            <p>The side of the source sign to copy text from</p>
     * @param clickedSide           <p>The side the player clicked to paste</p>
     * @param targetSign            <p>The sign to paste text onto</p>
     * @param player                <p>The player performing the copy paste</p>
     */
    private void copySignLine(@NotNull SignTextCopyRequest signTextCopyRequest,
                              @Nullable PlaceholderSign sourcePlaceholderSign, @NotNull Side sourceSide,
                              @NotNull Side clickedSide, @NotNull Sign targetSign, @NotNull Player player) {
        if (signTextCopyRequest.side() == null || signTextCopyRequest.sourceLine() == null) {
            throw new RuntimeException("Non-null variable is null");
        }

        int sourceLine = signTextCopyRequest.sourceLine();
        int destinationLine = signTextCopyRequest.destinationLine() != null ?
                signTextCopyRequest.destinationLine() : sourceLine;

        String text = signTextCopyRequest.sign().getSide(signTextCopyRequest.side()).getLine(sourceLine);
        if (sourcePlaceholderSign != null) {
            Map<Side, Map<Integer, String>> placeholders = sourcePlaceholderSign.getPlaceholders();
            if (placeholders.containsKey(sourceSide) && placeholders.get(sourceSide).containsKey(sourceLine)) {
                text = placeholders.get(sourceSide).get(sourceLine);
            }
        }

        doSignChange(targetSign, destinationLine, player, clickedSide, text);
    }

    /**
     * Copies sign text from one sign side to another
     *
     * @param signHandler           <p>The sign handler to use for dealing with placeholder signs</p>
     * @param sourcePlaceholderSign <p>The source sign's placeholder sign, or null</p>
     * @param targetPlaceholderSign <p>The destination's placeholder sign, or null</p>
     * @param sourceSide            <p>The side of the source sign to copy text from</p>
     * @param clickedSide           <p>The side the player clicked to paste</p>
     * @param sourceSign            <p>The sign to copy text from</p>
     * @param targetSign            <p>The sign to paste text onto</p>
     * @param player                <p>The player performing the copy paste</p>
     */
    private void copySignSideText(@NotNull PlaceholderSignHandler signHandler,
                                  @Nullable PlaceholderSign sourcePlaceholderSign,
                                  @Nullable PlaceholderSign targetPlaceholderSign, @NotNull Side sourceSide,
                                  @NotNull Side clickedSide, @NotNull Sign sourceSign, @NotNull Sign targetSign,
                                  @NotNull Player player) {
        // Remove old placeholders from the sign side
        Map<Integer, String> oldPlaceholders = null;
        if (targetPlaceholderSign != null) {
            oldPlaceholders = targetPlaceholderSign.getPlaceholders().remove(clickedSide);
        }
        if (pasteSignSide(sourceSide, clickedSide, sourceSign, targetSign,
                sourcePlaceholderSign, player) && targetPlaceholderSign != null) {
            // Restore the old placeholders if the sign change didn't finish
            targetPlaceholderSign.getPlaceholders().put(clickedSide, oldPlaceholders);
        }

        // Save any placeholder changes
        signHandler.save();

        // Apply changes
        targetSign.update();

        PlaceholderSigns.getInstance().getStringFormatter().displaySuccessMessage(player,
                PlaceholderSignMessage.SUCCESS_SIGN_PASTED);
    }

    /**
     * Copies sign text from one sign to another
     *
     * @param signHandler           <p>The sign handler to use for dealing with placeholder signs</p>
     * @param sourcePlaceholderSign <p>The source sign's placeholder sign, or null</p>
     * @param targetPlaceholderSign <p>The destination's placeholder sign, or null</p>
     * @param clickedSide           <p>The side the player clicked to paste</p>
     * @param sourceSign            <p>The sign to copy text from</p>
     * @param targetSign            <p>The sign to paste text onto</p>
     * @param player                <p>The player performing the copy paste</p>
     */
    private void copySignText(@NotNull PlaceholderSignHandler signHandler,
                              @Nullable PlaceholderSign sourcePlaceholderSign,
                              @Nullable PlaceholderSign targetPlaceholderSign, @NotNull Side clickedSide,
                              @NotNull Sign sourceSign, @NotNull Sign targetSign, @NotNull Player player) {
        // Un-register the old placeholder sign to remove its data
        if (targetPlaceholderSign != null) {
            signHandler.unregisterSign(targetPlaceholderSign);
        }

        // Copy both sides of the sign. Paste the front to whichever side the player is currently looking at
        if ((pasteSignSide(Side.FRONT, clickedSide, sourceSign, targetSign, sourcePlaceholderSign, player) ||
                pasteSignSide(Side.BACK, getOpposite(clickedSide), sourceSign, targetSign, sourcePlaceholderSign, player)) &&
                targetPlaceholderSign != null) {
            // Re-register the old placeholder sign if the new text couldn't be applied
            signHandler.registerSign(targetPlaceholderSign);
        }

        // Save any placeholder changes
        signHandler.save();

        // Apply changes
        targetSign.update();

        PlaceholderSigns.getInstance().getStringFormatter().displaySuccessMessage(player,
                PlaceholderSignMessage.SUCCESS_SIGN_PASTED);
    }


    /**
     * Checks if the player is trying to edit/modify a waxed sign without permission
     *
     * @param sign   <p>The sign the player is trying to edit</p>
     * @param player <p>The player trying to edit the sign</p>
     * @return <p>True if the player is trying to edit a waxed sign without permission</p>
     */
    private boolean checkWaxEdit(@NotNull Sign sign, @NotNull Player player, @NotNull String permissionNode) {
        if (sign.isWaxed() && !player.hasPermission(permissionNode)) {
            PlaceholderSigns.getInstance().getStringFormatter().displayErrorMessage(player,
                    PlaceholderSignMessage.ERROR_WAXED_NO_PERMISSION);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Pastes the data from the source sign to the target sign
     *
     * @param sourceSign <p>The source sign to copy data from</p>
     * @param targetSign <p>The target sign to paste to</p>
     * @param player     <p>The player pasting the sign</p>
     */
    private void pasteSign(@NotNull Sign sourceSign, @NotNull Sign targetSign, @NotNull Player player) {
        StringFormatter stringFormatter = PlaceholderSigns.getInstance().getStringFormatter();
        PlaceholderSign placeholderSign = PlaceholderSigns.getInstance().getSignHandler().getFromLocation(
                sourceSign.getLocation());

        // Get the lines, with placeholders replaced, and the sign change event processed
        if (pasteSignSide(Side.FRONT, Side.FRONT, sourceSign, targetSign, placeholderSign, player) ||
                pasteSignSide(Side.BACK, Side.BACK, sourceSign, targetSign, placeholderSign, player)) {
            return;
        }

        // Copy options
        copySignOptions(sourceSign, targetSign);

        // Apply changes
        targetSign.update();

        stringFormatter.displaySuccessMessage(player, PlaceholderSignMessage.SUCCESS_SIGN_PASTED);
    }

    /**
     * Pastes one side of a sign onto another
     *
     * @param sourceSide      <p>The sign side to copy</p>
     * @param destinationSide <p>The sign side to paste to</p>
     * @param source          <p>The source sign</p>
     * @param target          <p>The destination sign</p>
     * @param placeholderSign <p>The placeholder sign belonging to the source sign, or null</p>
     * @param player          <p>The player performing the paste</p>
     * @return <p>False if the pasting completed successfully. True if blocked by another plugin.</p>
     */
    private boolean pasteSignSide(@NotNull Side sourceSide, @NotNull Side destinationSide, @NotNull Sign source,
                                  @NotNull Sign target, @Nullable PlaceholderSign placeholderSign, @NotNull Player player) {
        SignSide sourceSignSide = source.getSide(sourceSide);
        SignSide targetSignSide = target.getSide(destinationSide);
        String[] frontLines = getFinalLines(target, sourceSignSide, sourceSide, destinationSide, placeholderSign, player);
        if (frontLines == null) {
            PlaceholderSigns.getInstance().getStringFormatter().displayErrorMessage(player,
                    PlaceholderSignMessage.ERROR_CANCELLED_BY_PROTECTION);
            return true;
        }
        for (int i = 0; i < frontLines.length; i++) {
            targetSignSide.setLine(i, ColorHelper.translateColorCodes(frontLines[i], ColorConversion.RGB));
        }
        return false;
    }

    /**
     * Gets the final lines from a sign side, after inserting placeholders and running a sign change event
     *
     * @param sign            <p>The sign that's changed</p>
     * @param signSide        <p>The side of the sign to get lines from</p>
     * @param sourceSide      <p>The side that text is taken from</p>
     * @param targetSide      <p>The side that text is changed for</p>
     * @param placeholderSign <p>The placeholder sign corresponding to the sign, if any</p>
     * @param player          <p>The player attempting to paste the sign</p>
     * @return <p>The final lines, or null if the event was cancelled</p>
     */
    @Nullable
    private String[] getFinalLines(@NotNull Sign sign, @NotNull SignSide signSide, @NotNull Side sourceSide,
                                   @NotNull Side targetSide, @Nullable PlaceholderSign placeholderSign,
                                   @NotNull Player player) {
        String[] sideLines = signSide.getLines();
        if (placeholderSign != null) {
            Map<Side, Map<Integer, String>> placeholders = placeholderSign.getPlaceholders();
            loadPlaceholders(sourceSide, sideLines, placeholders);
        }

        // Run the sign change event to allow protection plugins to cancel, and allow the sign text listener to trigger
        SignChangeEvent changeEvent = new SignChangeEvent(Objects.requireNonNull(sign.getBlock()), player, sideLines, targetSide);
        Bukkit.getPluginManager().callEvent(changeEvent);
        if (changeEvent.isCancelled()) {
            return null;
        }

        return changeEvent.getLines();
    }

    /**
     * Copies all sign option (dye, glowing, waxed) from the source to the target
     *
     * @param source <p>The source sign</p>
     * @param target <p>The target sign</p>
     */
    private void copySignOptions(@NotNull Sign source, @NotNull Sign target) {
        SignSide sourceFront = source.getSide(Side.FRONT);
        SignSide sourceBack = source.getSide(Side.BACK);
        SignSide targetFront = target.getSide(Side.FRONT);
        SignSide targetBack = target.getSide(Side.BACK);

        // Copy glowing state
        targetFront.setGlowingText(sourceFront.isGlowingText());
        targetBack.setGlowingText(sourceBack.isGlowingText());

        // Copy applied dye
        targetFront.setColor(sourceFront.getColor());
        targetBack.setColor(sourceBack.getColor());

        // Apply waxed state
        target.setWaxed(source.isWaxed());
    }

    /**
     * Loads placeholders from one side of a sign
     *
     * @param side         <p>The side to load from</p>
     * @param lines        <p>The array to write placeholders to</p>
     * @param placeholders <p>The placeholders to load</p>
     */
    private void loadPlaceholders(@NotNull Side side, @NotNull String[] lines,
                                  @NotNull Map<Side, Map<Integer, String>> placeholders) {
        if (!placeholders.containsKey(side)) {
            return;
        }

        Map<Integer, String> sidePlaceholders = placeholders.get(side);
        for (int i = 0; i < 4; i++) {
            if (sidePlaceholders.containsKey(i)) {
                lines[i] = sidePlaceholders.get(i);
            }
        }
    }

    /**
     * Alters the text on a sign, taking care of placeholders and sign change event triggers
     *
     * @param targetSign      <p>The sign to be changed</p>
     * @param destinationLine <p>The line to be changed on the sign</p>
     * @param player          <p>The player trying to change the sign</p>
     * @param targetSide      <p>The side of the sign to change the line for</p>
     * @param newText         <p>The new text to apply to the sign</p>
     */
    private void doSignChange(@NotNull Sign targetSign, int destinationLine, @NotNull Player player,
                              @NotNull Side targetSide, @NotNull String newText) {
        SignSide targetSignSide = targetSign.getSide(targetSide);
        String[] lines = targetSignSide.getLines();

        PlaceholderSignHandler signHandler = PlaceholderSigns.getInstance().getSignHandler();
        PlaceholderSign targetPlaceholderSign = PlaceholderSigns.getInstance().getSignHandler().getFromLocation(targetSign.getLocation());
        String oldPlaceholder = null;
        if (targetPlaceholderSign != null) {
            // Remove the old placeholder
            Map<Integer, String> placeholders = targetPlaceholderSign.getPlaceholders().get(targetSide);
            if (placeholders != null) {
                oldPlaceholder = placeholders.remove(destinationLine);
            }
        }

        lines[destinationLine] = newText;

        // Run the sign change event to allow protection plugins to cancel, and allow the sign text listener to trigger
        SignChangeEvent changeEvent = new SignChangeEvent(targetSign.getBlock(), player, lines, targetSide);
        Bukkit.getPluginManager().callEvent(changeEvent);
        if (changeEvent.isCancelled()) {
            if (targetPlaceholderSign != null) {
                // Restore the old placeholder if the action didn't complete
                targetPlaceholderSign.getPlaceholders().get(targetSide).put(destinationLine, oldPlaceholder);
            }
            PlaceholderSigns.getInstance().getStringFormatter().displayErrorMessage(player,
                    PlaceholderSignMessage.ERROR_CANCELLED_BY_PROTECTION);
            return;
        }

        // Update the sign with the new text
        targetSignSide.setLine(destinationLine, ColorHelper.translateColorCodes(changeEvent.getLines()[destinationLine],
                ColorConversion.RGB));
        targetSign.update();

        // Save any placeholder changes
        signHandler.save();

        PlaceholderSigns.getInstance().getStringFormatter().displaySuccessMessage(player,
                PlaceholderSignMessage.SUCCESS_SIGN_CHANGED);
    }

}
